/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: indexpara.C,v 1.1 1996/02/19 23:50:29 kilin Exp $ 
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <stdlib.h>
#include "TVdefine.h"
#include "TVconstants.h"
#include "TVindexpara.h"
#include "TVutil.h"

TVTree_Para::TVTree_Para()
{
   page_size = DEFAULT_PAGESIZE;
   unfold_dim = SIG_DIM;
   min_fill_percent = MIN_FILL_PERCENT;
   reinsert_percent = REINSERT_PERCENT;
   pickbranch_compare_count = DEFAULT_PICKBRANCH_COMPARE_COUNT;
}

TVTree_Para::TVTree_Para(const int pgsize, const int ud, const float mfp, const float rip, const int
pbcp)
{
   page_size = pgsize;
   unfold_dim = ud;
   min_fill_percent = mfp;
   reinsert_percent = rip;
   pickbranch_compare_count = pbcp;
}

TVTree_Para::TVTree_Para(const TVTree_Para& ip)
{
   page_size = ip.page_size;
   unfold_dim = ip.unfold_dim;
   min_fill_percent = ip.min_fill_percent;
   reinsert_percent = ip.reinsert_percent;
   pickbranch_compare_count = ip.pickbranch_compare_count;
}


TVTree_Para& TVTree_Para::operator=(const TVTree_Para& ip)
{
   page_size = ip.page_size;
   unfold_dim = ip.unfold_dim;
   min_fill_percent = ip.min_fill_percent;
   reinsert_percent = ip.reinsert_percent;
   pickbranch_compare_count = ip.pickbranch_compare_count;
   return *this;
}

void TVTree_Para::SetPageSize(int ps)
{
  page_size = ps;
}

void TVTree_Para::SetUnfoldDim(int ud)
{
  unfold_dim = ud;
}

void TVTree_Para::SetMinFillPercent(float mfp)
{
   min_fill_percent = mfp;
}

void TVTree_Para::SetReinsertPercent(float rip)
{
    reinsert_percent = rip;
}

void TVTree_Para::SetPickTVBranchCompareCount(int pbcp)
{
    pickbranch_compare_count = pbcp;
}

//  befriended functions

ostream& operator<< (ostream& os, const TVTree_Para& ip)
{
   os << "Page size : " << ip.page_size << "  Unfold dim : " <<  ip.unfold_dim << "  Min_fill % : " <<  ip.min_fill_percent << "  Reinsert % : " << ip.reinsert_percent << "  Pickbranch compare count : " << ip.pickbranch_compare_count;
   return os;
}


istream& operator>>(istream& is, TVTree_Para& ip)
{
   cout << "Input page_size : ";
   is >> ip.page_size;
   cout << "Input unfold_dim : ";
   is >> ip.unfold_dim;
   cout << "Input Min_fill % : ";
   is >> ip.min_fill_percent;
   cout << "Input Reinsert % : ";
   is >> ip.reinsert_percent;
   cout << "PickTVBranch compare count : ";
   is >> ip.pickbranch_compare_count;
   return is;
}



