/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: indexstat.h,v 1.1 1996/02/19 23:49:43 kilin Exp $ 


class TVTree_Stat {
    friend ostream& operator<< (ostream&, const TVTree_Stat&);
    friend ostream& operator<< (ostream&, const TVTree_Stat*&);
    friend istream& operator>> (istream&, TVTree_Stat&); 
    friend istream& operator>> (istream&, TVTree_Stat*&);

public :
    TVTree_Stat();    // Initialize a vector to be of no dimension
    TVTree_Stat(int); // new rectangle with center and radius
    TVTree_Stat(const TVTree_Stat&);

    ~TVTree_Stat();

    TVTree_Stat& operator=(const TVTree_Stat&);
     
    TVTree_Stat& IncrementPhase();
    TVTree_Stat& DummyPhase();
    TVTree_Stat& RestorePhase();
    TVTree_Stat& IncrementStat(int);
    TVTree_Stat& AddIntStat(int, int); 
    TVTree_Stat& AddFloatStat(int, float); 
    TVTree_Stat& SetZero(int);
    TVTree_Stat& SetPageSize(int);
    int GetIntStat(int);




private :  
    void Init_arrays();
    void Destroy_arrays();
    void Copy_arrays(const TVTree_Stat&);

    int max_phase;
    int current_phase;
    int stored_phase;  // store current phase when dummy is activated

    // no of disk read and write, assume memory buffer enough to hold intremediate results
    int *readcount, *writecount;    
    int *leafreadcount, *leafwritecount;    

    // no of times splits is invoked
    int *splitnodecount, *splitleafcount;

    // no of times reinsert is invoked
    int *reinsertcount;

    // calcualte the net ratio of split
    float *totalradiusratio;

    // number of reinsert branches
    int *insertbranchcount;

    // number of unfolding done
    int *unfoldcount;

    // number of folding back 
    int *foldbackcount;

    // number of element inserted
    int *insertelementcount;

    // info on tree
    int *treesize;
    int *nodecount;
    int *leafcount;

    // number of element distance calculation done
    int *worddiffcalcount;

    // other info for tree
    int pagesize;
};

