/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: linkedlist.C,v 1.2 1996/04/18 21:50:24 kilin Exp kilin $ 
#include <iostream.h>
#include <assert.h>
#include "TVconstants.h"
#include "TVdefine.h"
#include "TVvector.h"
#include "TVelement.h"
#include "TVlinkedlist.h"

LinkedListTVNode::LinkedListTVNode()
{
   ele = NULL;
   next = NULL;
}

LinkedListTVNode::~LinkedListTVNode()
{
}

Leaf_Element* LinkedListTVNode::Get()
{
   return ele;
}

LinkedListTVNode& LinkedListTVNode::Put(Leaf_Element* e)
{
  ele = e;
  return *this;
}


LinkedList::LinkedList()
{
   head = new LinkedListTVNode;
   head->next = NULL;
   tail = cur = NULL;
   count = 0;
}

LinkedList::~LinkedList()
{
   cur = head;
   while (cur->next)
     {
       cur = cur->next;
       delete head;
       head = cur;
     }
}

LinkedListTVNode* LinkedList::Head()
{
  return head->next;
}

LinkedListTVNode* LinkedList::Tail()
{
  return tail;
}

LinkedListTVNode* LinkedList::Cur()
{
  return cur;
}

LinkedListTVNode* LinkedList::Next()
{
  if (cur)
    cur = cur->next;
  return cur;
}

LinkedList& LinkedList::Append(Leaf_Element *e)
{
   if (!tail)
      tail = head;
   tail->next = new LinkedListTVNode;
   tail = tail->next;
   tail->Put(e);
   count++;
   return *this;
}

LinkedList& LinkedList::InsertHead(Leaf_Element *e)
{
   LinkedListTVNode *newnode = new LinkedListTVNode;
   newnode->Put(e);
   newnode->next = head->next;
   head->next = newnode;
   count++;
   return *this;
}

LinkedListTVNode* LinkedList::RemoveHead()
{
   LinkedListTVNode* res = head->next;

   if (res)
      {
         head = head->next;
	 if (tail == res)
	    tail = NULL;
         count--;
      }

  return res; 
}


LinkedList& LinkedList::ClearList()
{
   LinkedListTVNode *cur1;
   cur1 = cur = head;
   while (cur->next)
     {
       cur = cur->next;
       delete cur1;
       cur1 = cur;
     }
   cur = tail = NULL;
   head->next = NULL;
   return *this;
}


LinkedList& LinkedList::ResetToHead()
{
  cur = head->next;
  return *this;
}


LinkedList& LinkedList::ResetToTail()
{
  cur = tail;
  return *this;
}

int LinkedList::IsEmpty() const
{
   return (head->next == NULL);
}

int LinkedList::Count() const
{
   return count;
}

/*
ostream& LinkedList::PrintList(ostream &os, void (*PrintData)(ostream&, char*))
{
   int count = 1;
   os << "{";
   if (head->next)
     {
	PrintData(os, head->next->GetData());

	LinkedListTVNode *curnode = head->next->next;
	while (curnode)
	  {
	     if (!(count++ % 10))
		os << endl;
             os << " --> ";
	     PrintData(os, curnode->GetData());
	     curnode = curnode->next;
	  }
     }
  os << "}";
  return os;
}
*/
