/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: split2.C,v 1.3 1996/04/18 21:50:24 kilin Exp kilin $ 
// Please change sig_dim to active_dim

#include <iostream.h>
#include <assert.h>
#include <stdlib.h>
#include <fstream.h>
#include "TVdefine.h"
#include "TVconstants.h"
#include "TVvector.h"
#include "TVrectangle.h"
#include "TVelement.h"
#include "TVsimbuf.h"
#include "TVnode.h"
#include "TVutil.h"
#include "TVsplitcond.h"


SplitCond::SplitCond()
{
  index = -1;
}

SplitCond::SplitCond(const TVRectangle& lower, const TVRectangle& upper, int ind)
{
   radius = lower.GetRadius() + upper.GetRadius();
   int d1 = lower.GetCenterDim();
   int d2 = upper.GetCenterDim();
   if (d1 == d2)
      {
          diff =  lower.GetCenter().Distance(upper.GetCenter());
      }
   else
      {
          if (d1 < d2)
	    {
               diff =  lower.GetCenter().Distance(upper.GetCenter().ProjectFront(d1));
	    }
	 else
            {
               diff =  upper.GetCenter().Distance(lower.GetCenter().ProjectFront(d2));
            }

      }
  dim = d1 + d2;
  nointersectdim =  lower.NoIntersectDim(upper);
  index = ind;
}

SplitCond::SplitCond(const SplitCond& sc)
{
  radius = sc.radius;
  diff = sc.diff;
  dim = sc.dim;
  nointersectdim = sc.nointersectdim;
  index = sc.index; 
}

SplitCond& SplitCond::operator=(const SplitCond& sc)
{
  radius = sc.radius;
  diff = sc.diff;
  dim = sc.dim;
  nointersectdim = sc.nointersectdim;
  index = sc.index; 

  return *this;
}

int SplitCond::operator>(const SplitCond& sc) const
{
/*
   if (nointersectdim > sc.nointersectdim)
       return TRUE; 
   if (nointersectdim < sc.nointersectdim)
       return FALSE; 
*/
   if (dim > sc.dim)
       return TRUE; 
   if (dim < sc.dim)
       return FALSE; 
   return (radius < sc.radius) || ((radius - sc.radius < 1e-8) &&  (diff >= sc.diff));

}

int SplitCond::operator>=(const SplitCond& sc) const
{
/*
   if (nointersectdim > sc.nointersectdim)
       return TRUE; 
   if (nointersectdim < sc.nointersectdim)
       return FALSE; 
*/
   if (dim > sc.dim)
       return TRUE; 
   if (dim < sc.dim)
       return FALSE; 
   return  (radius < sc.radius) || ((radius - sc.radius < 1e-8) &&  (diff > sc.diff));

}

int SplitCond::IsEmpty() const
{
  return index == -1;
}
