/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: util.C,v 1.4 1996/04/18 21:50:24 kilin Exp kilin $ 
#include <stdlib.h>
#include <iostream.h>
#include <assert.h>
#include "TVdefine.h"

int sgn(float d1)
{
  if (d1 > 0)
     return(1);
  if (d1 < 0)
     return(-1);
  return(0);
}

/* get the value of the Hadmnard matrix */
/* row, column start at 0 */
int hadamard1(int n, int row, int column)
{

  int quad;

  if (n == 1)
     {
       if ((row > 1) || (column > 1))
          {
            cout << "Error in hadamard";
            assert(0);
          }
       else
          {
            if ((row == 1) && (column == 1))
               return(-1);
            else
               return(1);
          }
     }
  else
     {
       quad = 2 * (row >= (n / 2) ) + (column >= (n / 2));
       switch (quad){
         case 0 : return(hadamard1(n/2, row, column));
                  break;
         case 1 : return(hadamard1(n/2, row, column - (n/2)));
                  break;
         case 2 : return(hadamard1(n/2, row - (n/2), column));
                  break;
         case 3 : return(-hadamard1(n/2, row - (n/2), column - (n/2)));
                  break;
           }
     }

  return 0;
}

int hadamard2(int n, int row, int column)
{

  int quad;
  int n1, r1, c1;
  int ressig;

  n1 = n;
  r1 = row;
  c1 = column;
  ressig = 1;
  while (n1 > 2)
    {
       quad = 2 * (r1 >= (n1 / 2) ) + (c1 >= (n1 / 2));
       switch (quad){
         case 0 : break;
         case 1 : c1 = c1 - (n1/2);
                  break;
         case 2 : r1 = r1 - (n1/2);
                  break;
         case 3 : r1 = r1 - (n1/2);
		  c1 = c1 - (n1/2);
		  ressig = -ressig;
                  break;
           }
       n1 = n1/2;
     }
  if ((r1 > 1) || (c1 > 1))
     {
	printf("Ooooooops!\n");
     }
  else
     {
       if ((r1 == 1) && (c1 == 1))
          return(-ressig);
       else
          return(ressig);
     }
  return 0;
}

int hadamard(int n, int row, int column)
{
   int pof2 = 1, i;


   for (i = 0; i < n; i++)
       pof2 = pof2 * 2;
   return(hadamard2(pof2, row, column));
}

/* return the corresponding entry in the smallest hadamard matrix with at least
minsize rows */
int hadamard_minsize(int minsize, int row, int column)
{
   int pof2 = 1, i;


   for (pof2 = 1; pof2 < minsize; pof2 = pof2 * 2)
       ;
   return(hadamard2(pof2, row, column));
}

/* get the logrithaithm */
int logplusone(int n)
{
  int sol = 0;

  while (n > 0)
    {
       n = n / 2;
       sol++;
    }
  return(sol);
}



/*
// for hilbert curve calculation

void inleave(char xcoor[], char ycoor[], char bitstr[])
{       int xnum,ynum,ptr;
        xnum = strlen(xcoor);
        ynum = strlen(ycoor);
        if (xnum >= ynum)
                ptr = 2 * xnum;
        else ptr = 2 * ynum;
        bitstr[ptr--] = '\0';
        while ((xnum != 0) && (ynum != 0))
                {
                bitstr[ptr--] = ycoor[--ynum];
                bitstr[ptr--] = xcoor[--xnum];
                }
        if (xnum != 0)
                {
                while (xnum != 0)
                        {
                        bitstr[ptr--] = '0';
                        bitstr[ptr--] = xcoor[--xnum];
                        }
                }
        else if (ynum != 0)
                {
                while (ynum != 0)
                        {
                        bitstr[ptr--] = ycoor[--ynum];
                        bitstr[ptr--] = '0';
                        }
                }
}
*/

//*******************************************************************/
//   CONVERT :  Takes a binary number, divides it into 2-bit        */
//   strings, and assigns a decimal value to each string.           */
//   '00' gets 0; '01' gets 1; '10' gets 3, and '11' gets 2.        */
//*******************************************************************/
/*
void convert(char bitstr[], int string[] ,int count)
{       int n,i;
        n=0;
        i=0;
        while (i < count)
                {
                if (bitstr[n++] == '0')
                        { 
                        if (bitstr[n] == '0')
                                string[i++] = 0; 
                        else if (bitstr[n] == '1')
                                string[i++] = 1; 
                        }
                else      // the first number is one 
                        { 
                        if (bitstr[n] == '0')
                                string[i++] = 3; 
                        else if (bitstr[n] == '1')
                                string[i++] = 2; 
                        }
                n = n + 1;
                }
}
*/

//*******************************************************************/
//   SWITCHES :  Switches numbers to correspond with the rotation   */
//   of Hilbert's curve.                                            */
//*******************************************************************/
/*
void switches(int string[], int count)
{       int i,j;
        for (i=0;i<count;i++)
                {
                if (string[i] == 0)
                        {
                        for (j=i+1;j<count;j++)
                                {
                                if (string[j] == 1)
                                        string[j] = 3;
                                else if (string[j] == 3)
                                        string[j] = 1;
                                }
                        }
                else if (string[i] == 3)
                        {
                        for (j=i+1;j<count;j++)
                                {
                                if (string[j] == 0)
                                        string[j] = 2;
                                else if (string[j] == 2)
                                        string[j] = 0;
                                }
                        }
                }        
}
*/

//*******************************************************************/
//   COMPUTE :  Computes the zvalue, given an array of manipulated  */
//   numbers from SWITCHES.                                         */
//*******************************************************************/
/*
int compute(int string[], int count)
{       int i,n;
        int zvalue;
        n=4;
        zvalue = string[count-1];
        for (i=count-2; i>=0; i--)
                {
                zvalue = zvalue + (string[i] * n);
                n = n * 4;
                }
        return(zvalue);
}
*/

//*******************************************************************/
//   ZCOMP :  Introduction procedure to calculate the zvalue of     */
//   the Hilbert curve.                                             */
//*******************************************************************/
/*
int zcomp(char xcoor[], char ycoor[])
{
        char bitstr[40];
        int count;
        int string[20];
        int zvalue;
        inleave(xcoor,ycoor,bitstr);
        count = (strlen(bitstr)) / 2;
        convert(bitstr,string,count);
        switches(string,count);
        zvalue = compute(string,count);
        return(zvalue);
}
*/



//*******************************************************************/
//  NCOMP :  Introduction procedure to calculate the zvalue of the  */
//  curve.                                                          */
//*******************************************************************/
/*
int ncomp(char xcoor[], char ycoor[])
{
        char bstring[40];
        int nvalue,n,i,count;
        inleave(xcoor,ycoor,bstring);
        count = strlen(bstring);
        if (bstring[count-1] == '0')
                nvalue = 0;
        else nvalue = 1;
        n = 2;
        for (i=count-2; i>=0; i--)
                {
                if (bstring[i] == '1')
                        nvalue = nvalue + n;
                n = n * 2;
                }
        return(nvalue);
}
*/

//*******************************************************************/
//   ADDONE :  Adds one to a binary number                          */
//*******************************************************************/
/*
void addone(char coor[], char newbin[])
{
        int count;
        count = strlen(coor);
        newbin[count--] = '\0';
        while ((coor[count] == '1') && (count > 0))
                newbin[count--] = '0';
        if (coor[count] == '0')
                {
                newbin[count--] = '1';
                while (count >= 0)
                        {
                        newbin[count] = coor[count];
                        count--;
                        }
                }
        else if (count == 0)
                newbin[count] = 'x';
}
*/

//*******************************************************************/
//   SUBONE :  Subtracts one from a binary number                   */
//*******************************************************************/
/*
void subone(char coor[], char newbin[])
{
        int count;
        count = strlen(coor);
        newbin[count--] = '\0';
        while ((coor[count] == '0') && (count > 0))
                newbin[count--] = '1';
        if (coor[count] == '1')
                {
                newbin[count--] = '0';
                while (count >= 0)
                        {
                        newbin[count] = coor[count];
                        count--;
                        }
                }
        else if (count == 0)
                newbin[count] = 'x';
}
*/


//*******************************************************************/
//   ICONVERT :  Takes a binary number, divides it into 2-bit       */
//   strings, and assigns a decimal value to each string.           */
//   '00' gets 0; '01' gets 1; '10' gets 3, and '11' gets 2.        */
//*******************************************************************/
/*
void iconvert(int zcoors[], int string[], int count)
{       int t,n,i;
        zcoors[0] = 0;
        zcoors[1] = 0;
        t = 1;
        for (i=count-1;i>=0;i--)
                {
                if (string[i] == 1)
                        zcoors[1] = zcoors[1] + t;
                else if (string[i] == 2)
                        {
                        zcoors[0] = zcoors[0] + t;
                        zcoors[1] = zcoors[1] + t;
                        }
                else if (string[i] == 3)
                        {
                        zcoors[0] = zcoors[0] + t;
                        }
                t = t * 10;
                }        
}
*/

//*******************************************************************/
//   ISWITCHES :  Switches numbers to correspond with the rotation  */
//   of Hilbert's curve.                                            */
//*******************************************************************/
/*
void iswitches(int string[], int count)
{       int i,j;
        for (i=count-1;i>=0;i--)
                {
                if (string[i] == 0)
                        {
                        for(j=i+1;j<count;j++)
                                {
                                if (string[j] == 1)
                                        string[j] = 3;
                                else if (string[j] == 3)
                                        string[j] = 1;
                                }
                        }
                else if (string[i] == 3)
                        {
                        for (j=i+1;j<count;j++)
                                {
                                if (string[j] == 0)
                                        string[j] = 2;
                                else if (string[j] == 2)
                                        string[j] = 0;
                                }
                        }
                }
}
*/

//*******************************************************************/
//   ICOMPUTE :  Computes the zvalue, given an array of manipulated */
//   numbers from SWITCHES.                                         */
//*******************************************************************/
/*
void icompute(int zvalue,int string[],int n,int degree)
{       int i,count;
        if (zvalue >= (n*n))
                string[0] = 9999;
        else
                {
                count = (n*n)/4;
                for (i=0;i<degree;i++)
                        {
                        string[i] = zvalue/count;
                        zvalue = zvalue % count;
                        count = count/4;
                        }
                }
}
*/


//*******************************************************************/
//   ZINVERSE  Introduction procedure to calculate the zvalue of    */
//   the Hilbert curve.                                             */
//*******************************************************************/
/*
void zinverse(int zvalue,int n,int zcoors[])
{
        char xcoor[10],ycoor[10];
        int count,i;
        int string[20];
        count = 0;
        i=n;
        while (i > 1)
                {
                i = i / 2;
                count++;
                }
        icompute(zvalue,string,n,count);
        if (string[0] != 9999)
                {
                iswitches(string,count);
                iconvert(zcoors,string,count);
                }
        else zcoors[0] = 9999;
}
*/

//*******************************************************************/
//*******************************************************************/
/*
void ninverse(int nvalue,int coors[])
{
        int i,y,x,n;
        i = 1;
        coors[0] = 0;
        coors[1] = 0;
        while (nvalue > 0)
                {
                y = nvalue % 2;
                coors[1] = coors[1] + (y*i);
                nvalue = nvalue / 2;
                x = nvalue % 2;
                coors[0] = coors[0] + (x*i);
                nvalue = nvalue / 2;
                i = i * 10;
                }
}



int binary(int decvalue)
{
        int i,x,binval;
        i = 1;
        binval = 0;
        while (decvalue > 0)
                {
                x = decvalue % 2;
                binval = binval + (x*i);
                decvalue = decvalue / 2;
                i = i * 10;
                }
        return(binval);
}


void binaryvec(int decvalue,char *zcoor, int degree)
//convert decimal number decvalue into binary number stored in char array zcoor
{int i,z[20];
 for(i=degree-1;i>=0;i--)
  {z[i]=decvalue%2;
   decvalue=decvalue/2;
  };
 for (i=0;i<degree;i++)
  if (z[i]==0)
    zcoor[i]='0';
  else zcoor[i]='1';
 zcoor[degree]='\0';
}

// n=span of coordinate
int hvalue(int x, int  y, int n)
//return the hilbert curve value for coordinate (x,y)
{char xcoor[20],ycoor[20];
 int i,degree, pow2;
 i=n; degree=0, pow2 = 1;
 while (i!=1)
  {i=i/2;
   degree++;
   pow2 = pow2 * 2;
  };
 if (n != pow2)
    degree++;
 binaryvec(x,xcoor,degree);
 binaryvec(y,ycoor,degree);
 return(zcomp(xcoor,ycoor));
}
*/


int equal(float x, float y)
{
   return ((abs(x - y) <= PRECISION * x) && (abs(y-x) <= PRECISION * y));
}

int equal(double x, double y)
{
   return ((abs(x - y) <= PRECISION * x) && (abs(y-x) <= PRECISION * y));
}


// Check if alow..blow intersect with ahigh..bhigh
// Return TRUE even if only touching
int intersect(float alow, float ahigh, float blow, float bhigh)
{
   int result = TRUE;
   if (alow < blow)
      result = (blow <= ahigh);
   else if (alow > blow)
      result = (alow <= bhigh);
   return result;
}
