/* 
 
                  COPYRIGHT NOTICE
 
This material was developed by Christos Faloutsos and King-Ip Lin
at the University of Maryland, College Park, Department of Computer Science.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings.
 
1. Any copy made of this software must include this copyright notice
in full.
 
2. All materials developed as a consequence of the use of this
software shall duly acknowledge such use, in accordance with the usual
standards of acknowledging credit in academic research.
 
3. The authors have made no warranty or representation that the
operation of this software will be error-free or suitable for any
application, and they are under under no obligation to provide any
services, by way of maintenance, update, or otherwise.  The software
is an experimental prototype offered on an as-is basis.
 
4. Redistribution for profit requires the express, written permission
of the authors.
 
*/

// Author : $Author$
// Date : $Date$
// Id : $Id$
// $Id: indextest.C,v 1.6 1996/04/18 21:50:24 kilin Exp kilin $ 
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <stdlib.h>
#include <String.h>
#include "TVdefine.h"
#include "TVconstants.h"
#include "TVvector.h"
#include "TVelement.h"
#include "TVrectangle.h"
#include "TVsimbuf.h"
#include "TVnode.h"
#include "TVindexstat.h"
#include "TVindexpara.h"
#include "TVindex.h"
#include "TVutil.h"
#include "dataadd.h"
#include "indextest.h"

extern "C" {
 char* strcpy(char*, const char*);
 int strcmp(const char*, const char*);
//  unsigned int strlen(const char*);

}

#define DEBUG 1

Storage *globstorage;

int eletested;

TVector GetTVector(DataPoints& dp, int dim = 0)
{

   dim = (dim ? dim : 30);
   TVector v(dim);
 
   for (int i = 0; i < dim; i++)
       v[i] = dp[i];
   return v;
}

TestPara::TestPara() : ip(default_pagesize, default_sig_dim, default_min_fill_percent, default_reinsert_percent, default_pickbranch_compare_count)
{
   optionfromfile = FALSE;
   strcpy(osfilename, default_osfilename);

   loadindex = FALSE;
   strcpy(lifilename, default_lifilename);

   performinsert = TRUE;
   performsearch = TRUE;

   numinput= default_numinput;
   inskip= default_inskip;
   numsearch= default_numsearch;
   searchskip= default_searchskip;
   searchskiphead= default_searchskiphead;
   searchdist= default_searchdist;

   searchcode = searchall;
   searchoption = searchexact;

   searchprint = FALSE;
   searchprintfile = NULL;

   strcpy(ifilename, default_ifilename);
   strcpy(sfilename, default_sfilename);

   saveindex = FALSE;

   strcpy(svfilename, default_svfilename);

   printindex = FALSE;
   onedataperline = TRUE;
   checkvalid = FALSE;
   buffersize = default_buffersize;
   disktime = default_disktime;

   quicksearch = FALSE;
}

void ChangeOptions(int argc, char *argv[], TestPara& tp)
{
   for (int i  = 2; i <= argc; i++)
       {
        if (!strcmp(argv[i-1], "-optf"))              // option file
            {
               tp.optionfromfile = TRUE;
               strcpy(tp.osfilename, argv[i++]);
               i = argc + 1;
            }
        else if (!strcmp(argv[i-1], "-li"))                // insert or search
            {
               tp.loadindex =  TRUE;
	       tp.performinsert = FALSE;
               strcpy(tp.lifilename, argv[i++]);
            }
        else if (!strcmp(argv[i-1], "-ps"))            // page size
                {
                   tp.ip.SetPageSize(atoi(argv[i++]));
                }
        else if (!strcmp(argv[i-1], "-ud"))            // unfold dim
                {
                   tp.ip.SetUnfoldDim(atoi(argv[i++]));
                }
        else if (!strcmp(argv[i-1], "-mfp"))            // min fill percent
                {
                   tp.ip.SetMinFillPercent(atof(argv[i++]));
                }
        else if (!strcmp(argv[i-1], "-rtp.ip"))            // reinsert percentage
                {
                   tp.ip.SetReinsertPercent(atof(argv[i++]));
                }
        else if (!strcmp(argv[i-1], "-pbcp"))            // pickbranch compare percentage
                {
                   tp.ip.SetPickTVBranchCompareCount(atoi(argv[i++]));
                }
        else if (!strcmp(argv[i-1], "-ni"))            // no insertion
                {
                   tp.performinsert = FALSE;
                }
        else if (!strcmp(argv[i-1], "-if"))            // insert file name
                {
                   strcpy(tp.ifilename, argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-ic"))            // input count
                {
                   tp.performinsert = TRUE;
                   tp.numinput = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-isc"))            // input skip count
                {
                   tp.inskip  = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-ns"))            // no search
                {
                   tp.performsearch = FALSE;
                }
        else if (!strcmp(argv[i-1], "-sf"))            // search file name
                {
                   strcpy(tp.sfilename, argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-sc"))            // search count
                {
                   tp.numsearch = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-ssc"))            // search skip count
                {
                   tp.searchskip  = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-sshc"))            // search skip head
                {
                   tp.searchskiphead = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-se"))            // search exist
                {
                   tp.searchcode = searchexist; 
                }
        else if (!strcmp(argv[i-1], "-sap"))            // search approx rectangle
                {
                   tp.searchoption = searchappdia;
                   tp.searchdist = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-saq"))            // search approx rectangle
                {
                   tp.searchoption = searchappsq;
                   tp.searchdist = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-sed"))            // search approx word
                {
                   tp.searchoption = searcheditdist;
                   tp.searchdist = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-sedns"))            // search approx word
                {
                   tp.searchoption = searcheditdistnosub;
                   tp.searchdist = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-sp"))            //  print all found elements
                {
                   tp.searchprint = TRUE;
                   if ((i < argc) && (argv[i][0] != '-'))
		      {
			tp.searchprintfile = new char[strlen(argv[i]) + 1];
			strcpy(tp.searchprintfile, argv[i++]);
		      }	
                }
        else if (!strcmp(argv[i-1], "-nsv"))            // no save index
                {
                  tp.saveindex = FALSE;
                }
        else if (!strcmp(argv[i-1], "-svi"))            // save file name
                {
                   tp.saveindex = TRUE;
                   if ((i < argc) && (argv[i][0] != '-'))
                      strcpy(tp.svfilename, argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-pi"))            // print index
                {
                  tp.printindex = TRUE;
                }
        else if (!strcmp(argv[i-1], "-pin"))            // print index
                {
                  tp.printindex = TRUE;
                  tp.onedataperline = FALSE;
                }
        else if (!strcmp(argv[i-1], "-cv"))            //  check validity
                {
                  tp.checkvalid = TRUE;
                }
        else if (!strcmp(argv[i-1], "-dd"))            // set dimension of data 
                {
                   DataPoints::dim = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-bs"))            // input count
                {
                   tp.buffersize = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-dt"))            // input count
                {
                   tp.disktime = atoi(argv[i++]);
                }
        else if (!strcmp(argv[i-1], "-qs"))            // no save index
                {
                  tp.quicksearch = TRUE;
                }
       }
}

void PrintOptions(TestPara& tp)
{
  if (tp.loadindex)
     cout << "Index read from file : " << tp.lifilename << "\n\n";
  else
    {
     cout << "Index built from scratch.\n";
     cout << "Index parameters :\n";
     cout << tp.ip;
     cout << "\n\n";
    }

  if (tp.performinsert)
     {
       cout << "Insert from file " << tp.ifilename << "\n";
       cout << "Number of elements inserted " << tp.numinput << "\n";
       cout << "Skip between element inserted :  " << tp.inskip << "\n";
     }
  else
     cout << "No insertion performed\n";
  cout << endl << endl;

  if (tp.performsearch)
     {
       switch (tp.searchoption) {
         case searchexact :  cout << "Exact search performed";
                             break;
         case searchappdia :  cout << "Approximate search rectangle region with radius " << tp.searchdist;
                              break;
         case searchappdis :  cout << "Approximate search with distance (no filtering) " << tp.searchdist;
                             break;
         case searcheditdist :  cout << "Approximate search with editing distance (do filtering) " << tp.searchdist;
                             break;
         default :
		   break; 
         }
       cout << "\n";
       cout << "Search file name " << tp.sfilename << "\n";
       cout << "Number of elements searched : " << tp.numsearch << "\n";
       cout << "Skip between element : " << tp.searchskip << "\n";
       cout << "Skip from head of file : " << tp.searchskiphead << "\n";
     }
  else
     cout << "No insertion performed\n";
  cout << endl << endl;

  if (tp.saveindex)
     cout << "Index saved at file " << tp.svfilename << "\n";
  else
     cout << "Index not saved\n";
  cout << endl;


}

main(int argc, char *argv[])
{
  TestPara tp;
  TVTree *ind1;


  eletested = 0;

  // Change options

  ChangeOptions(argc, argv, tp);

  // Print options selected

  PrintOptions(tp);

  // Creat Storage

  globstorage = new Storage(tp.ip.page_size, tp.buffersize);

  // Create index

  if (tp.loadindex)
     ind1 = new TVTree(tp.lifilename, ReadData, GetFeature);
  else
     ind1 = new TVTree(tp.ip, PHASES);
    
#ifdef DEBUG
  cout << "\nEnd of create index\n";
#endif

// cout << "Count : " << ind1->Count() << endl;

  if (tp.performinsert)
     {
        ifstream ifile(tp.ifilename, ios::in);

        // inserting data
        for (int icount = 0 ; icount < tp.numinput; icount++)
          {
/*
	     DataPoints dp;
	     ifile >> dp;
	     size = dp.Size();
	     char *newdata = dp();
	     Leaf_Element e(newdata, size);
*/

	     DataPoints *dp = new DataPoints;
	     ifile >> *dp;
	     int size = dp->Size();

/*
if (icount == 3542)
 cout << icount << "  Hello\n";
*/

// cout << "icount : " << icount << "  Count : " << ind1->Count() << endl;

	     Leaf_Element e((char *)dp, size);
            


             ind1->InsertElement(e, GetFeature);
	     DataPoints dump;
             for (int skipcount = 0; skipcount < tp.inskip; skipcount++)
	        ifile >> dump;

//	     delete [] newdata;
          }
        ifile.close();


     }

#ifdef DEBUG
   cout << "end of insert\n";
#endif




   cout << "Build buffer result : " << *globstorage <<  "   Time for read,write : " << (globstorage->diskfetchcount + globstorage->pagewritecount) * tp.disktime / 1000.0 << " seconds"<< endl;

   if (tp.printindex)
       ind1->TVTreePrint(cout, PrintData, tp.onedataperline);

   globstorage->emptybuffer();
   globstorage->initstats();

   if (tp.checkvalid)
      {
	cout << "Checking validity : ";
        if (ind1->ValidityCheck())
	   cout << "good";
        else
	   cout << "ERROR!!!";
	cout << "   Count : " << ind1->Count() << endl;
      }



   // Search
   if (tp.performsearch)
      {
	ind1->IncrementPhase(); // new phase for gathering search stats
	int size;
        ifstream sfile(tp.sfilename, ios::in);
        int hitcount = 0;
        globstorage->emptybuffer();
        globstorage->initstats();

        ostream *search_print_out;
        ofstream search_print_out_file;


        // Set output for search results
        if (tp.searchprint)
           {
             if (tp.searchprintfile)
                {
                   search_print_out_file.open(tp.searchprintfile, ios::out);
                   search_print_out = &search_print_out_file;
                }
             else
                search_print_out = &cout;
           }
        else
          search_print_out = NULL;

        int i;

        // Skip the top of the file
        for (i = 0; i < tp.searchskiphead; i++)
	   {
	   DataPoints dp1;
	   sfile >> dp1;
	   }


	for (int scount = 0; scount < tp.numsearch; scount++)
            {
	        DataPoints dp;
                sfile >> dp;
		int curhitcount;

		// Searching
		switch (tp.searchoption) {
		  case searchexact : {
		                        if (tp.searchprint)	
		                           (*search_print_out) << "\n -- " << dp << " : ";		
					  char **res;
	                                int size = dp.Size();
					curhitcount = ind1->Search((char *)&dp, size, res, GetKey, GetFeature, Equal, tp.searchcode);
					hitcount += curhitcount;
				        if (tp.searchprint)	
					   {
						for (int spp = 0; spp < curhitcount;)
						  {
						  	(*search_print_out) << res[spp++] << " / ";
							if (!(spp % 10))
						           (*search_print_out) << endl;
						  }
						(*search_print_out) << endl;
					   }
				     }
				     break;
		   case searchappdia : {

					  TVector v = GetTVector(dp, 3);
					  float lower[3], upper[3];
					  for (int j = 0; j < 3; j++)	
					   {
						lower[j] = v[j] - tp.searchdist;
						upper[j] = v[j] + tp.searchdist;
					   }
		
				
		                          if (tp.searchprint)	
					     {
		                              (*search_print_out) << "\n -- " << dp  << " {(" << lower[0];		
					      int pv;
					      for (pv = 1; pv < 3; pv++)
						 (*search_print_out) << ", " << lower[pv];
		                              (*search_print_out) << "), (" << upper[0];		
					      for (pv = 1; pv < 3; pv++)
						 (*search_print_out) << ", " << upper[pv];
		                              (*search_print_out) << ")} : ";		
					     }
					  char **res;
	                                  int size = dp.Size();
					  curhitcount = ind1->Search((char *)&dp, size, 3, lower, upper,  res, GetKey, GetFeature, tp.searchcode);
					  hitcount += curhitcount;
		                          (*search_print_out) << " Count : " << curhitcount << " ";
				          if (tp.searchprint)	
					     {
						for (int spp = 0; spp < curhitcount;)
						  {
						  	(*search_print_out) << res[spp++] << " / ";
							if (!(spp % 10))
						           (*search_print_out) << endl;
						  }
						(*search_print_out) << endl;
					     }
					}
				      break;
		   case searcheditdist : {
		                        if (tp.searchprint)	
		                           (*search_print_out) << "\n -- " << dp << " : ";		
					char **res;
	                                int size = dp.Size();
					curhitcount = ind1->Search(tp.searchdist, (char *)&dp, size,  res, GetKey, GetFeature, Distance, tp.searchcode);
					hitcount += curhitcount;
				        if (tp.searchprint)	
					   {
						for (int spp = 0; spp < curhitcount;)
						  {
						  	(*search_print_out) << res[spp++] << " / ";
							if (!(spp % 10))
						           (*search_print_out) << endl;
						  }
						(*search_print_out) << endl;
					   }
				       }
		   case searchappdis : {
		                        if (tp.searchprint)	
		                           (*search_print_out) << "\n -- " << dp << " : ";		
					char **res;
	                                int size = dp.Size();
					curhitcount = ind1->Search(tp.searchdist, (char *)&dp, size,  res, GetKey, GetFeature, tp.searchcode);
					hitcount += curhitcount;
				        if (tp.searchprint)	
					   {
						for (int spp = 0; spp < curhitcount;)
						  {
						  	(*search_print_out) << res[spp++] << " / ";
							if (!(spp % 10))
						           (*search_print_out) << endl;
						  }
						(*search_print_out) << endl;
					   }
				       }
		   default : 
			     break;

		  }
            
	       // skipping elements
               for (i = 0; i < tp.searchskip; i++)
                sfile >> dp;

// cout << scount << " " << *globstorage << endl;
            }

	sfile.close();
        cout << endl << "Total hits : " << hitcount << endl;

        cout << "Search result : " << *globstorage <<  "   Time for read,write : " << (globstorage->diskfetchcount + globstorage-> pagewritecount) * tp.disktime / 1000.0 << " seconds"<< endl;
        cout << "Element tested : " << eletested << endl;
      }

#ifdef DEBUG
   cout << "end of search\n";
#endif


   // Print stats
  ind1->GatherTreeStats();
  ind1->PrintStats(cout);

  if (tp.saveindex)
    ind1->SaveTVTree(tp.svfilename, WriteData);
  
  ind1->FreeTree();
  delete ind1;
}

